]> git.sesse.net Git - stockfish/blobdiff - src/tt.h
Don't allocate MAX_THREADS hash tables if not necessary
[stockfish] / src / tt.h
index bd6aae99b9e2918a4b008e28af9abc2f17e75f4e..69cdad0c17dc19f19d351a36e0b79faf5e696700 100644 (file)
--- a/src/tt.h
+++ b/src/tt.h
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-
 #if !defined(TT_H_INCLUDED)
 #define TT_H_INCLUDED
 
-////
-//// Includes
-////
-
-#include "depth.h"
-#include "position.h"
-#include "value.h"
+#include <iostream>
 
+#include "move.h"
+#include "types.h"
 
-////
-//// Types
-////
 
-/// The TTEntry class is the class of transposition table entries
+/// The TTEntry is the class of transposition table entries
 ///
 /// A TTEntry needs 128 bits to be stored
 ///
 /// bit 64-79: value
 /// bit 80-95: depth
 /// bit 96-111: static value
/// bit 112-127: king danger value
 ///
 /// the 32 bits of the data field are so defined
 ///
-/// bit  0-16: move
-/// bit 17-19: not used
-/// bit 20-22: value type
+/// bit  0-15: move
+/// bit 16-20: not used
+/// bit 21-22: value type
 /// bit 23-31: generation
 
 class TTEntry {
 
 public:
-  void save(uint32_t k, Value v, ValueType t, Depth d, Move m, int g, Value statV, Value kd) {
-
-      key32 = k;
-      data = (m & 0x1FFFF) | (t << 20) | (g << 23);
-      value16     = int16_t(v);
-      depth16     = int16_t(d);
-      staticValue = int16_t(statV);
-      kingDanger  = int16_t(kd);
+  void save(uint32_t k, Value v, ValueType t, Depth d, Move m, int g, Value statV, Value statM) {
+
+    key32        = (uint32_t)k;
+    move16       = (uint16_t)m;
+    valueType    = (uint8_t)t;
+    generation8  = (uint8_t)g;
+    value16      = (int16_t)v;
+    depth16      = (int16_t)d;
+    staticValue  = (int16_t)statV;
+    staticMargin = (int16_t)statM;
   }
+  void set_generation(int g) { generation8 = (uint8_t)g; }
 
-  uint32_t key() const { return key32; }
-  Depth depth() const { return Depth(depth16); }
-  Move move() const { return Move(data & 0x1FFFF); }
-  Value value() const { return Value(value16); }
-  ValueType type() const { return ValueType((data >> 20) & 7); }
-  int generation() const { return data >> 23; }
-  Value static_value() const { return Value(staticValue); }
-  Value king_danger() const { return Value(kingDanger); }
+  uint32_t key() const              { return key32; }
+  Depth depth() const               { return (Depth)depth16; }
+  Move move() const                 { return (Move)move16; }
+  Value value() const               { return (Value)value16; }
+  ValueType type() const            { return (ValueType)valueType; }
+  int generation() const            { return (int)generation8; }
+  Value static_value() const        { return (Value)staticValue; }
+  Value static_value_margin() const { return (Value)staticMargin; }
 
 private:
   uint32_t key32;
-  uint32_t data;
-  int16_t value16;
-  int16_t depth16;
-  int16_t staticValue;
-  int16_t kingDanger;
+  uint16_t move16;
+  uint8_t valueType, generation8;
+  int16_t value16, depth16, staticValue, staticMargin;
 };
 
 
-/// This is the number of TTEntry slots for each position
+/// This is the number of TTEntry slots for each cluster
 const int ClusterSize = 4;
 
-/// Each group of ClusterSize number of TTEntry form a TTCluster
-/// that is indexed by a single position key. TTCluster size must
-///  be not bigger then a cache line size, in case it is less then
-/// it should be padded to guarantee always aligned accesses.
+
+/// TTCluster consists of ClusterSize number of TTEntries. Size of TTCluster
+/// must not be bigger than a cache line size. In case it is less, it should
+/// be padded to guarantee always aligned accesses.
 
 struct TTCluster {
   TTEntry data[ClusterSize];
 };
 
 
-/// The transposition table class. This is basically just a huge array
-/// containing TTEntry objects, and a few methods for writing new entries
-/// and reading new ones.
+/// The transposition table class. This is basically just a huge array containing
+/// TTCluster objects, and a few methods for writing and reading entries.
 
 class TranspositionTable {
 
+  TranspositionTable(const TranspositionTable&);
+  TranspositionTable& operator=(const TranspositionTable&);
+
 public:
   TranspositionTable();
   ~TranspositionTable();
@@ -111,26 +106,68 @@ public:
   void store(const Key posKey, Value v, ValueType type, Depth d, Move m, Value statV, Value kingD);
   TTEntry* retrieve(const Key posKey) const;
   void new_search();
-  void insert_pv(const Position& pos, Move pv[]);
-  void extract_pv(const Position& pos, Move bestMove, Move pv[], const int PLY_MAX);
   TTEntry* first_entry(const Key posKey) const;
+  void refresh(const TTEntry* tte) const;
 
 private:
   size_t size;
   TTCluster* entries;
-  uint8_t generation;
+  uint8_t generation; // Size must be not bigger then TTEntry::generation8
 };
 
 extern TranspositionTable TT;
 
 
-/// TranspositionTable::first_entry returns a pointer to the first
-/// entry of a cluster given a position. The low 32 bits of the key
-/// are used to get the index in the table.
+/// TranspositionTable::first_entry() returns a pointer to the first entry of
+/// a cluster given a position. The lowest order bits of the key are used to
+/// get the index of the cluster.
 
 inline TTEntry* TranspositionTable::first_entry(const Key posKey) const {
 
-  return entries[uint32_t(posKey) & (size - 1)].data;
+  return entries[((uint32_t)posKey) & (size - 1)].data;
 }
 
+
+/// TranspositionTable::refresh() updates the 'generation' value of the TTEntry
+/// to avoid aging. Normally called after a TT hit.
+
+inline void TranspositionTable::refresh(const TTEntry* tte) const {
+
+  const_cast<TTEntry*>(tte)->set_generation(generation);
+}
+
+
+/// A simple fixed size hash table used to store pawns and material
+/// configurations. It is basically just an array of Entry objects.
+/// Without cluster concept or overwrite policy.
+
+template<class Entry, int HashSize>
+struct SimpleHash {
+
+  typedef SimpleHash<Entry, HashSize> Base;
+
+  void init() {
+
+    if (entries)
+        return;
+
+    entries = new (std::nothrow) Entry[HashSize];
+    if (!entries)
+    {
+        std::cerr << "Failed to allocate " << HashSize * sizeof(Entry)
+                  << " bytes for hash table." << std::endl;
+        exit(EXIT_FAILURE);
+    }
+    memset(entries, 0, HashSize * sizeof(Entry));
+  }
+
+  virtual ~SimpleHash() { delete [] entries; }
+
+  Entry* find(Key key) const { return entries + ((uint32_t)key & (HashSize - 1)); }
+  void prefetch(Key key) const { ::prefetch((char*)find(key)); }
+
+protected:
+  Entry* entries;
+};
+
 #endif // !defined(TT_H_INCLUDED)
@@
 /// bit 80-95: depth
 /// bit 96-111: static value
-/// bit 112-127: king danger value
+/// bit 112-127: margin of static value
 ///
 /// the 32 bits of the data field are so defined